/*
// Magnet Maker Deluxe!
// (c) 2005 nathan vander wilt
// for PHYS-203L, Dr. Allen, Dordt College, Sioux Center, IA
//
// gcc -std=c99 -fast magnet2.c -o magnet2
*/

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <math.h>
#include <string.h>

typedef unsigned char uchar;
typedef unsigned int uint;
#define true 1
#define false 0

#define MAX_ITS 500
//#define DISPLAY


struct magnet {
 int w;
 int h;
 uchar* grid;
};

inline uint mag_index(uint x, uint y, uint w) {
 return y*w+x;
}

struct magnet mag_create(uint w, uint h) {
// init pseudorandom seed
 //srand( time(NULL) );   // use if the line below breaks
 srandomdev();

  
 struct magnet mag;
 mag.w=w;
 mag.h=h;
 mag.grid=malloc(w*h);
 
 for (int i=0; i<w*h; i++) {
  mag.grid[i]=random()%2;
  //mag.grid[i]=i;
 }

 return mag;
}

uint mag_pbc(int x, int w) {
 if ( x>-1 && x<w) return x;
 if ( x<0 ) return (w+(x%w))%w; // TBD: simplify?
 if ( x>(w-1) ) return (x%w);
}

inline uchar mag_cell_value(uint x, uint y, struct magnet mag) {
 int _x=mag_pbc(x,mag.w);
 int _y=mag_pbc(y,mag.h);

 return mag.grid[mag_index(_x,_y,mag.w)];
}

uchar mag_around_cell(uint x, uint y, struct magnet mag) {
 uchar total=0;
 
 for (int i=-1; i<=1; i++) {
  for (int j=-1; j<=1; j++) total+=mag_cell_value(x+i,y+j,mag);
 }
 
 return total;
}

struct magnet mag_step(struct magnet in) {
 struct magnet out;
 memcpy(&out,&in,sizeof(struct magnet));
 out.grid=malloc(out.w*out.h);
 
 for (int i=0; i<out.h; i++) {
  for (int j=0; j<out.w; j++) {
   out.grid[mag_index(j,i,out.w)]= ( mag_around_cell(j,i,in) > 4) ? 1 : 0;
  }
 }

 return out;
}

void mag_release(struct magnet done) {
 if (done.grid) free(done.grid);
 memset(&done,0,sizeof(struct magnet));
}

uchar mag_equal(struct magnet mag1, struct magnet mag2) {
 if ( mag1.w != mag2.w ) return false;
 if ( mag1.h != mag2.h ) return false;

 if ( memcmp(mag1.grid, mag2.grid, mag1.w*mag2.h) ) return false;
 else return true;
}


// Tests

void test_rand() {
 int j=0;
 int top=10000000;

// init pseudorandom seed
// srand( time(NULL) );   // use if the line below breaks
 srandomdev();

 for (int i=0; i<top; i++) {
  j+=random()&01;
 }
 fprintf(stderr,"%1.5f", j / (top-1.0) );
}

void test_pbc() {
 for (int i=-20; i<25; i++) {
  printf("%i -> %i\n",i,mag_pbc(i,10));
 }
}

// End Tests


float measure_domain(struct magnet mag) {
 uint total=0;
 
 for (int i=0; i<mag.h; i++) {
  for (int j=0; j<mag.w; j++) {
   int num=mag_around_cell(j,i,mag);
   if ( !(num%9) ) ++total;
  }
 }
 
 return (float)total/(mag.w*mag.h);
}

float measure_magnet(struct magnet mag) {
 int total=0;
 for (int i=0; i<mag.w*mag.h; i++) total+=mag.grid[i];

 int area=mag.w*mag.h;
 
 return fabs(2.0*total-area)/area;
}

float measure_up(struct magnet mag) {
 int total=0;
 for (int i=0; i<mag.w*mag.h; i++) total+=mag.grid[i];
 
 return (float)total/(mag.w*mag.h);
}


void print_magnet(struct magnet mag) {
 for (uint i=0; i<mag.h; i++) {
  for (uint j=0; j<mag.w; j++) {
   printf(" %i", mag_cell_value(j,i,mag) );
  }
  printf("\n");
 }
 printf("Domain size:    %3.2f%%\n", 100.0*measure_domain(mag) );
 printf("Magnetization:  %3.2f%%\n", 100.0*measure_magnet(mag) );
 printf("Percent ones:   %3.2f%%\n", 100.0*measure_up(mag) );
}


void do_run(uint x, uint y, FILE* output) {
 uint its=1;
 struct magnet new;
 struct magnet older;
 memset(&older,0,sizeof(struct magnet));
 memset(&new,0,sizeof(struct magnet));
 
 struct magnet old=mag_create(x,y);
#ifdef DISPLAY
 print_magnet(old);
 printf("\n\n");
#endif

 time_t start=time(NULL);
 
 while (its<MAX_ITS) {
  if (output) {
   fprintf(output,"%i, %i, %i, %3.2f, %3.2f, %3.2f\n",old.w,old.h,its,100.0*measure_domain(old),100.0*measure_magnet(old),100.0*measure_up(old));
   fflush(output);
  }
  new=mag_step(old);
  if ( mag_equal(new,old) ) break;
  if ( mag_equal(new,older) ) break;

#ifdef DISPLAY  
  print_magnet(new);
  printf("\n\n");
#endif

  mag_release(older);
  older=old;
  old=new;
  ++its;
 }
 time_t end=time(NULL);
 
 if (output) {
  fprintf(output,"%i, %i, %i, %3.2f, %3.2f, %3.2f, %i, *\n",new.w,new.h,its,100.0*measure_domain(new),100.0*measure_magnet(new),100.0*measure_up(new),end-start);
  fflush(output);
 }
#ifdef DISPLAY
 printf("%i iterations in ~%i seconds\n",its,end-start);
#endif
 mag_release(older);
 mag_release(old);
 if(its<MAX_ITS) mag_release(new);
}

int main() {
 fprintf(stderr,"Magnet Maker Deluxe!, version ");
 test_rand();
 fprintf(stderr,"\n"); 
 fprintf(stderr,"(c) 2005, nathan vander wilt\n");

 uint tests[]={ 1,10, 10,50, 25,250, 50,500, 100,1000, 500,5000, 1000,10000, 5000,50000, 10000,100000, 10000,1000000 };
 uint times=20;
 uint size=1;

 for ( uint step=0; step<sizeof(tests)/sizeof(uint); step+=2 ) {
  while ( size<tests[step+1] ) {
   fprintf(stderr,"Computing size: %ix%i",size,size);
   for (uint i=0; i<times; i++) { fprintf(stderr,"."); fflush(stderr); do_run(size,size,stdout); }
   fprintf(stderr,"\n");
   size+=tests[step];
  }
 }
}
